@@ -19,11 +19,6 @@ from subprocess import PIPE
 
 from .deprecation import RemovedInNextVersionWarning
 
-if False:
-    # For type annotation
-    from typing import Any  # NOQA
-
-
 # by default, all DeprecationWarning under sphinx package will be emit.
 # Users can avoid this by using environment variable: PYTHONWARNINGS=
 if 'PYTHONWARNINGS' not in os.environ:
@@ -32,8 +27,8 @@ if 'PYTHONWARNINGS' not in os.environ:
 warnings.filterwarnings('ignore', "'U' mode is deprecated",
                         DeprecationWarning, module='docutils.io')
 
-__version__ = '3.5.2'
-__released__ = '3.5.2'  # used when Sphinx builds its own docs
+__version__ = '4.0.0'
+__released__ = '4.0.0'  # used when Sphinx builds its own docs
 
 #: Version info for better programmatic use.
 #:
@@ -43,7 +38,7 @@ __released__ = '3.5.2'  # used when Sphinx builds its own docs
 #:
 #: .. versionadded:: 1.2
 #:    Before version 1.2, check the string ``sphinx.__version__``.
-version_info = (3, 5, 2, 'final', 0)
+version_info = (4, 0, 0, 'final', 0)
 
 package_dir = path.abspath(path.dirname(__file__))
 
@@ -57,8 +52,8 @@ if __version__.endswith('+'):
     try:
         ret = subprocess.run(['git', 'show', '-s', '--pretty=format:%h'],
                              cwd=package_dir,
-                             stdout=PIPE, stderr=PIPE)
+                             stdout=PIPE, stderr=PIPE, encoding='ascii')
         if ret.stdout:
-            __display_version__ += '/' + ret.stdout.decode('ascii').strip()
+            __display_version__ += '/' + ret.stdout.strip()
     except Exception:
         pass
