@@ -13,9 +13,9 @@ import os
 import posixpath
 import re
 import sys
-import warnings
+from datetime import datetime
 from os import path
-from typing import IO, Any, Dict, Iterable, Iterator, List, Set, Tuple
+from typing import IO, Any, Dict, Iterable, Iterator, List, Set, Tuple, Type
 from urllib.parse import quote
 
 from docutils import nodes
@@ -29,7 +29,6 @@ from sphinx import __display_version__, package_dir
 from sphinx.application import Sphinx
 from sphinx.builders import Builder
 from sphinx.config import ENUM, Config
-from sphinx.deprecation import RemovedInSphinx40Warning
 from sphinx.domains import Domain, Index, IndexEntry
 from sphinx.environment.adapters.asset import ImageAdapter
 from sphinx.environment.adapters.indexentries import IndexEntries
@@ -39,7 +38,7 @@ from sphinx.highlighting import PygmentsBridge
 from sphinx.locale import _, __
 from sphinx.search import js_index
 from sphinx.theming import HTMLThemeFactory
-from sphinx.util import logging, md5, progress_message, status_iterator
+from sphinx.util import isurl, logging, md5, progress_message, status_iterator
 from sphinx.util.docutils import is_html5_writer_available, new_document
 from sphinx.util.fileutil import copy_asset
 from sphinx.util.i18n import format_date
@@ -49,11 +48,6 @@ from sphinx.util.osutil import copyfile, ensuredir, os_path, relative_uri
 from sphinx.util.tags import Tags
 from sphinx.writers.html import HTMLTranslator, HTMLWriter
 
-if False:
-    # For type annotation
-    from typing import Type  # for python3.5.1
-
-
 # HTML5 Writer is available or not
 if is_html5_writer_available():
     from sphinx.writers.html5 import HTML5Translator
@@ -88,9 +82,9 @@ class Stylesheet(str):
     its filename (str).
     """
 
-    attributes = None   # type: Dict[str, str]
-    filename = None     # type: str
-    priority = None     # type: int
+    attributes: Dict[str, str] = None
+    filename: str = None
+    priority: int = None
 
     def __new__(cls, filename: str, *args: str, priority: int = 500, **attributes: Any
                 ) -> "Stylesheet":
@@ -114,9 +108,9 @@ class JavaScript(str):
     its filename (str).
     """
 
-    attributes = None   # type: Dict[str, str]
-    filename = None     # type: str
-    priority = None     # type: int
+    attributes: Dict[str, str] = None
+    filename: str = None
+    priority: int = None
 
     def __new__(cls, filename: str, priority: int = 500, **attributes: str) -> "JavaScript":
         self = str.__new__(cls, filename)
@@ -185,7 +179,7 @@ class StandaloneHTMLBuilder(Builder):
     allow_parallel = True
     out_suffix = '.html'
     link_suffix = '.html'  # defaults to matching out_suffix
-    indexer_format = js_index  # type: Any
+    indexer_format: Any = js_index
     indexer_dumps_unicode = True
     # create links to original images from images [True/False]
     html_scaled_image_link = True
@@ -201,26 +195,26 @@ class StandaloneHTMLBuilder(Builder):
     use_index = False
     download_support = True  # enable download role
 
-    imgpath = None          # type: str
-    domain_indices = []     # type: List[Tuple[str, Type[Index], List[Tuple[str, List[IndexEntry]]], bool]]  # NOQA
+    imgpath: str = None
+    domain_indices: List[Tuple[str, Type[Index], List[Tuple[str, List[IndexEntry]]], bool]] = []  # NOQA
 
     def __init__(self, app: Sphinx) -> None:
         super().__init__(app)
 
         # CSS files
-        self.css_files = []  # type: List[Dict[str, str]]
+        self.css_files: List[Dict[str, str]] = []
 
         # JS files
-        self.script_files = []  # type: List[JavaScript]
+        self.script_files: List[JavaScript] = []
 
     def init(self) -> None:
         self.build_info = self.create_build_info()
         # basename of images directory
         self.imagedir = '_images'
         # section numbers for headings in the currently visited document
-        self.secnumbers = {}  # type: Dict[str, Tuple[int, ...]]
+        self.secnumbers: Dict[str, Tuple[int, ...]] = {}
         # currently written docname
-        self.current_docname = None  # type: str
+        self.current_docname: str = None
 
         self.init_templates()
         self.init_highlighter()
@@ -256,6 +250,14 @@ class StandaloneHTMLBuilder(Builder):
                 return jsfile
         return None
 
+    def _get_style_filename(self) -> str:
+        if self.config.html_style is not None:
+            return self.config.html_style
+        elif self.theme:
+            return self.theme.get_config('theme', 'stylesheet')
+        else:
+            return 'default.css'
+
     def get_theme_config(self) -> Tuple[str, Dict]:
         return self.config.html_theme, self.config.html_theme_options
 
@@ -291,6 +293,9 @@ class StandaloneHTMLBuilder(Builder):
             self.dark_highlighter = None
 
     def init_css_files(self) -> None:
+        self.add_css_file('pygments.css', priority=200)
+        self.add_css_file(self._get_style_filename(), priority=200)
+
         for filename, attrs in self.app.registry.css_files:
             self.add_css_file(filename, **attrs)
 
@@ -305,6 +310,8 @@ class StandaloneHTMLBuilder(Builder):
         self.css_files.append(Stylesheet(filename, **kwargs))  # type: ignore
 
     def init_js_files(self) -> None:
+        self.add_js_file('documentation_options.js', id="documentation_options",
+                         data_url_root='', priority=200)
         self.add_js_file('jquery.js', priority=200)
         self.add_js_file('underscore.js', priority=200)
         self.add_js_file('doctools.js', priority=200)
@@ -326,7 +333,7 @@ class StandaloneHTMLBuilder(Builder):
         self.script_files.append(JavaScript(filename, **kwargs))
 
     @property
-    def default_translator_class(self) -> "Type[nodes.NodeVisitor]":  # type: ignore
+    def default_translator_class(self) -> Type[nodes.NodeVisitor]:  # type: ignore
         if not html5_ready or self.config.html4_writer:
             return HTMLTranslator
         else:
@@ -358,6 +365,7 @@ class StandaloneHTMLBuilder(Builder):
                 buildinfo = BuildInfo.load(fp)
 
             if self.build_info != buildinfo:
+                logger.debug('[build target] did not match: build_info ')
                 yield from self.env.found_docs
                 return
         except ValueError as exc:
@@ -372,6 +380,7 @@ class StandaloneHTMLBuilder(Builder):
             template_mtime = 0
         for docname in self.env.found_docs:
             if docname not in self.env.all_docs:
+                logger.debug('[build target] did not in env: %r', docname)
                 yield docname
                 continue
             targetname = self.get_outfilename(docname)
@@ -383,6 +392,14 @@ class StandaloneHTMLBuilder(Builder):
                 srcmtime = max(path.getmtime(self.env.doc2path(docname)),
                                template_mtime)
                 if srcmtime > targetmtime:
+                    logger.debug(
+                        '[build target] targetname %r(%s), template(%s), docname %r(%s)',
+                        targetname,
+                        datetime.utcfromtimestamp(targetmtime),
+                        datetime.utcfromtimestamp(template_mtime),
+                        docname,
+                        datetime.utcfromtimestamp(path.getmtime(self.env.doc2path(docname))),
+                    )
                     yield docname
             except OSError:
                 # source doesn't exist anymore
@@ -419,10 +436,10 @@ class StandaloneHTMLBuilder(Builder):
             self.load_indexer(docnames)
 
         self.docwriter = HTMLWriter(self)
-        self.docsettings = OptionParser(
+        self.docsettings: Any = OptionParser(
             defaults=self.env.settings,
             components=(self.docwriter,),
-            read_config_files=True).get_default_values()  # type: Any
+            read_config_files=True).get_default_values()
         self.docsettings.compact_lists = bool(self.config.html_compact_lists)
 
         # determine the additional indices to include
@@ -431,8 +448,7 @@ class StandaloneHTMLBuilder(Builder):
         indices_config = self.config.html_domain_indices
         if indices_config:
             for domain_name in sorted(self.env.domains):
-                domain = None  # type: Domain
-                domain = self.env.domains[domain_name]
+                domain: Domain = self.env.domains[domain_name]
                 for indexcls in domain.indices:
                     indexname = '%s-%s' % (domain.name, indexcls.name)
                     if isinstance(indices_config, list):
@@ -457,7 +473,7 @@ class StandaloneHTMLBuilder(Builder):
 
         self.relations = self.env.collect_relations()
 
-        rellinks = []  # type: List[Tuple[str, str, str, str]]
+        rellinks: List[Tuple[str, str, str, str]] = []
         if self.use_index:
             rellinks.append(('genindex', _('General Index'), 'I', _('index')))
         for indexname, indexcls, content, collapse in self.domain_indices:
@@ -470,13 +486,6 @@ class StandaloneHTMLBuilder(Builder):
         self._script_files = list(self.script_files)
         self._css_files = list(self.css_files)
 
-        if self.config.html_style is not None:
-            stylename = self.config.html_style
-        elif self.theme:
-            stylename = self.theme.get_config('theme', 'stylesheet')
-        else:
-            stylename = 'default.css'
-
         self.globalcontext = {
             'embedded': self.embedded,
             'project': self.config.project,
@@ -484,7 +493,8 @@ class StandaloneHTMLBuilder(Builder):
             'version': self.config.version,
             'last_updated': self.last_updated,
             'copyright': self.config.copyright,
-            'master_doc': self.config.master_doc,
+            'master_doc': self.config.root_doc,
+            'root_doc': self.config.root_doc,
             'use_opensearch': self.config.html_use_opensearch,
             'docstitle': self.config.html_title,
             'shorttitle': self.config.html_short_title,
@@ -499,7 +509,7 @@ class StandaloneHTMLBuilder(Builder):
             'language': self.config.language,
             'css_files': self.css_files,
             'sphinx_version': __display_version__,
-            'style': stylename,
+            'style': self._get_style_filename(),
             'rellinks': rellinks,
             'builder': self.name,
             'parents': [],
@@ -786,12 +796,12 @@ class StandaloneHTMLBuilder(Builder):
                        excluded, context=context, renderer=self.templates, onerror=onerror)
 
     def copy_html_logo(self) -> None:
-        if self.config.html_logo:
+        if self.config.html_logo and not isurl(self.config.html_logo):
             copy_asset(path.join(self.confdir, self.config.html_logo),
                        path.join(self.outdir, '_static'))
 
     def copy_html_favicon(self) -> None:
-        if self.config.html_favicon:
+        if self.config.html_favicon and not isurl(self.config.html_favicon):
             copy_asset(path.join(self.confdir, self.config.html_favicon),
                        path.join(self.outdir, '_static'))
 
@@ -889,21 +899,11 @@ class StandaloneHTMLBuilder(Builder):
         # only index pages with title
         if self.indexer is not None and title:
             filename = self.env.doc2path(pagename, base=None)
-            try:
-                metadata = self.env.metadata.get(pagename, {})
-                if 'nosearch' in metadata:
-                    self.indexer.feed(pagename, filename, '', new_document(''))
-                else:
-                    self.indexer.feed(pagename, filename, title, doctree)
-            except TypeError:
-                # fallback for old search-adapters
-                self.indexer.feed(pagename, title, doctree)  # type: ignore
-                indexer_name = self.indexer.__class__.__name__
-                warnings.warn(
-                    'The %s.feed() method signature is deprecated. Update to '
-                    '%s.feed(docname, filename, title, doctree).' % (
-                        indexer_name, indexer_name),
-                    RemovedInSphinx40Warning, stacklevel=2)
+            metadata = self.env.metadata.get(pagename, {})
+            if 'nosearch' in metadata:
+                self.indexer.feed(pagename, filename, '', new_document(''))
+            else:
+                self.indexer.feed(pagename, filename, title, doctree)
 
     def _get_local_toctree(self, docname: str, collapse: bool = True, **kwargs: Any) -> str:
         if 'includehidden' not in kwargs:
@@ -1026,7 +1026,7 @@ class StandaloneHTMLBuilder(Builder):
 
         # revert script_files and css_files
         self.script_files[:] = self._script_files
-        self.css_files[:] = self.css_files
+        self.css_files[:] = self._css_files
 
         self.update_page_context(pagename, templatename, ctx, event_arg)
         newtmpl = self.app.emit_firstresult('html-page-context', pagename,
@@ -1108,7 +1108,7 @@ class StandaloneHTMLBuilder(Builder):
 
 def convert_html_css_files(app: Sphinx, config: Config) -> None:
     """This converts string styled html_css_files to tuple styled one."""
-    html_css_files = []  # type: List[Tuple[str, Dict]]
+    html_css_files: List[Tuple[str, Dict]] = []
     for entry in config.html_css_files:
         if isinstance(entry, str):
             html_css_files.append((entry, {}))
@@ -1125,7 +1125,7 @@ def convert_html_css_files(app: Sphinx, config: Config) -> None:
 
 def convert_html_js_files(app: Sphinx, config: Config) -> None:
     """This converts string styled html_js_files to tuple styled one."""
-    html_js_files = []  # type: List[Tuple[str, Dict]]
+    html_js_files: List[Tuple[str, Dict]] = []
     for entry in config.html_js_files:
         if isinstance(entry, str):
             html_js_files.append((entry, {}))
@@ -1140,7 +1140,7 @@ def convert_html_js_files(app: Sphinx, config: Config) -> None:
     config.html_js_files = html_js_files  # type: ignore
 
 
-def setup_js_tag_helper(app: Sphinx, pagename: str, templatexname: str,
+def setup_js_tag_helper(app: Sphinx, pagename: str, templatename: str,
                         context: Dict, doctree: Node) -> None:
     """Set up js_tag() template helper.
 
@@ -1157,6 +1157,8 @@ def setup_js_tag_helper(app: Sphinx, pagename: str, templatexname: str,
                 if value is not None:
                     if key == 'body':
                         body = value
+                    elif key == 'data_url_root':
+                        attrs.append('data-url_root="%s"' % pathto('', resource=True))
                     else:
                         attrs.append('%s="%s"' % (key, html.escape(value, True)))
             if js.filename:
@@ -1164,11 +1166,35 @@ def setup_js_tag_helper(app: Sphinx, pagename: str, templatexname: str,
         else:
             # str value (old styled)
             attrs.append('src="%s"' % pathto(js, resource=True))
-        return '<script %s>%s</script>' % (' '.join(attrs), body)
+
+        if attrs:
+            return '<script %s>%s</script>' % (' '.join(attrs), body)
+        else:
+            return '<script>%s</script>' % body
 
     context['js_tag'] = js_tag
 
 
+def setup_resource_paths(app: Sphinx, pagename: str, templatename: str,
+                         context: Dict, doctree: Node) -> None:
+    """Set up relative resource paths."""
+    pathto = context.get('pathto')
+
+    # favicon_url
+    favicon = context.get('favicon')
+    if favicon and not isurl(favicon):
+        context['favicon_url'] = pathto('_static/' + favicon, resource=True)
+    else:
+        context['favicon_url'] = favicon
+
+    # logo_url
+    logo = context.get('logo')
+    if logo and not isurl(logo):
+        context['logo_url'] = pathto('_static/' + logo, resource=True)
+    else:
+        context['logo_url'] = logo
+
+
 def validate_math_renderer(app: Sphinx) -> None:
     if app.builder.format != 'html':
         return
@@ -1209,27 +1235,47 @@ def validate_html_static_path(app: Sphinx, config: Config) -> None:
 
 def validate_html_logo(app: Sphinx, config: Config) -> None:
     """Check html_logo setting."""
-    if config.html_logo and not path.isfile(path.join(app.confdir, config.html_logo)):
+    if (config.html_logo and
+            not path.isfile(path.join(app.confdir, config.html_logo)) and
+            not isurl(config.html_logo)):
         logger.warning(__('logo file %r does not exist'), config.html_logo)
         config.html_logo = None  # type: ignore
 
 
 def validate_html_favicon(app: Sphinx, config: Config) -> None:
     """Check html_favicon setting."""
-    if config.html_favicon and not path.isfile(path.join(app.confdir, config.html_favicon)):
+    if (config.html_favicon and
+            not path.isfile(path.join(app.confdir, config.html_favicon)) and
+            not isurl(config.html_favicon)):
         logger.warning(__('favicon file %r does not exist'), config.html_favicon)
         config.html_favicon = None  # type: ignore
 
 
+class _stable_repr_object():
+
+    def __repr__(self):
+        return '<object>'
+
+
+UNSET = _stable_repr_object()
+
+
 def migrate_html_add_permalinks(app: Sphinx, config: Config) -> None:
     """Migrate html_add_permalinks to html_permalinks*."""
-    if config.html_add_permalinks:
-        if (isinstance(config.html_add_permalinks, bool) and
-                config.html_add_permalinks is False):
-            config.html_permalinks = False  # type: ignore
-        else:
-            config.html_permalinks_icon = html.escape(config.html_add_permalinks)  # type: ignore  # NOQA
+    html_add_permalinks = config.html_add_permalinks
+    if html_add_permalinks is UNSET:
+        return
 
+    # RemovedInSphinx60Warning
+    logger.warning(__('html_add_permalinks has been deprecated since v3.5.0. '
+                      'Please use html_permalinks and html_permalinks_icon instead.'))
+    if not html_add_permalinks:
+        config.html_permalinks = False  # type: ignore[attr-defined]
+        return
+
+    config.html_permalinks_icon = html.escape(  # type: ignore[attr-defined]
+        html_add_permalinks
+    )
 
 # for compatibility
 import sphinxcontrib.serializinghtml  # NOQA
@@ -1261,7 +1307,7 @@ def setup(app: Sphinx) -> Dict[str, Any]:
     app.add_config_value('html_sidebars', {}, 'html')
     app.add_config_value('html_additional_pages', {}, 'html')
     app.add_config_value('html_domain_indices', True, 'html', [list])
-    app.add_config_value('html_add_permalinks', None, 'html')
+    app.add_config_value('html_add_permalinks', UNSET, 'html')
     app.add_config_value('html_permalinks', True, 'html')
     app.add_config_value('html_permalinks_icon', '¶', 'html')
     app.add_config_value('html_use_index', True, 'html')
@@ -1283,7 +1329,7 @@ def setup(app: Sphinx) -> Dict[str, Any]:
     app.add_config_value('html_search_scorer', '', None)
     app.add_config_value('html_scaled_image_link', True, 'html')
     app.add_config_value('html_baseurl', '', 'html')
-    app.add_config_value('html_codeblock_linenos_style', 'table', 'html',
+    app.add_config_value('html_codeblock_linenos_style', 'inline', 'html',  # RemovedInSphinx60Warning  # NOQA
                          ENUM('table', 'inline'))
     app.add_config_value('html_math_renderer', None, 'env')
     app.add_config_value('html4_writer', False, 'html')
@@ -1302,6 +1348,7 @@ def setup(app: Sphinx) -> Dict[str, Any]:
     app.connect('config-inited', validate_html_favicon, priority=800)
     app.connect('builder-inited', validate_math_renderer)
     app.connect('html-page-context', setup_js_tag_helper)
+    app.connect('html-page-context', setup_resource_paths)
 
     # load default math renderer
     app.setup_extension('sphinx.ext.mathjax')
