@@ -48,7 +48,7 @@
     resolved to a Python object, and otherwise it becomes simple emphasis.
     This can be used as the default role to make links 'smart'.
 
-    :copyright: Copyright 2007-2020 by the Sphinx team, see AUTHORS.
+    :copyright: Copyright 2007-2021 by the Sphinx team, see AUTHORS.
     :license: BSD, see LICENSE for details.
 """
 
@@ -60,46 +60,41 @@ import sys
 import warnings
 from os import path
 from types import ModuleType
-from typing import Any, Dict, List, Tuple
-from typing import cast
+from typing import Any, Dict, List, Optional, Tuple, Type, cast
 
 from docutils import nodes
 from docutils.nodes import Element, Node, system_message
 from docutils.parsers.rst import directives
-from docutils.parsers.rst.states import Inliner, RSTStateMachine, Struct, state_classes
+from docutils.parsers.rst.states import RSTStateMachine, Struct, state_classes
 from docutils.statemachine import StringList
 
 import sphinx
 from sphinx import addnodes
 from sphinx.application import Sphinx
-from sphinx.deprecation import RemovedInSphinx40Warning, RemovedInSphinx50Warning
+from sphinx.config import Config
+from sphinx.deprecation import RemovedInSphinx50Warning
 from sphinx.environment import BuildEnvironment
 from sphinx.environment.adapters.toctree import TocTree
-from sphinx.ext.autodoc import Documenter, INSTANCEATTR
+from sphinx.ext.autodoc import INSTANCEATTR, Documenter
 from sphinx.ext.autodoc.directive import DocumenterBridge, Options
 from sphinx.ext.autodoc.importer import import_module
 from sphinx.ext.autodoc.mock import mock
 from sphinx.locale import __
 from sphinx.pycode import ModuleAnalyzer, PycodeError
-from sphinx.util import rst, logging
-from sphinx.util.docutils import (
-    NullReporter, SphinxDirective, SphinxRole, new_document, switch_source_input
-)
+from sphinx.util import logging, rst
+from sphinx.util.docutils import (NullReporter, SphinxDirective, SphinxRole, new_document,
+                                  switch_source_input)
 from sphinx.util.matching import Matcher
+from sphinx.util.typing import OptionSpec
 from sphinx.writers.html import HTMLTranslator
 
-if False:
-    # For type annotation
-    from typing import Type  # for python3.5.1
-
-
 logger = logging.getLogger(__name__)
 
 
 periods_re = re.compile(r'\.(?:\s+)')
 literal_re = re.compile(r'::\s*$')
 
-WELL_KNOWN_ABBREVIATIONS = (' i.e.',)
+WELL_KNOWN_ABBREVIATIONS = ('et al.', ' i.e.',)
 
 
 # -- autosummary_toc node ------------------------------------------------------
@@ -170,18 +165,20 @@ def autosummary_table_visit_html(self: HTMLTranslator, node: autosummary_table)
 # -- autodoc integration -------------------------------------------------------
 
 # current application object (used in `get_documenter()`).
-_app = None  # type: Sphinx
+_app: Sphinx = None
 
 
 class FakeDirective(DocumenterBridge):
     def __init__(self) -> None:
         settings = Struct(tab_width=8)
         document = Struct(settings=settings)
+        env = BuildEnvironment()
+        env.config = Config()
         state = Struct(document=document)
-        super().__init__({}, None, Options(), 0, state)  # type: ignore
+        super().__init__(env, None, Options(), 0, state)
 
 
-def get_documenter(app: Sphinx, obj: Any, parent: Any) -> "Type[Documenter]":
+def get_documenter(app: Sphinx, obj: Any, parent: Any) -> Type[Documenter]:
     """Get an autodoc.Documenter class suitable for documenting the given
     object.
 
@@ -229,7 +226,7 @@ class Autosummary(SphinxDirective):
     optional_arguments = 0
     final_argument_whitespace = False
     has_content = True
-    option_spec = {
+    option_spec: OptionSpec = {
         'caption': directives.unchanged_required,
         'toctree': directives.unchanged,
         'nosignatures': directives.flag,
@@ -314,7 +311,7 @@ class Autosummary(SphinxDirective):
         """
         prefixes = get_import_prefixes_from_env(self.env)
 
-        items = []  # type: List[Tuple[str, str, str, str]]
+        items: List[Tuple[str, str, str, str]] = []
 
         max_item_chars = 50
 
@@ -434,29 +431,6 @@ class Autosummary(SphinxDirective):
 
         return [table_spec, table]
 
-    def warn(self, msg: str) -> None:
-        warnings.warn('Autosummary.warn() is deprecated',
-                      RemovedInSphinx40Warning, stacklevel=2)
-        logger.warning(msg)
-
-    @property
-    def genopt(self) -> Options:
-        warnings.warn('Autosummary.genopt is deprecated',
-                      RemovedInSphinx40Warning, stacklevel=2)
-        return self.bridge.genopt
-
-    @property
-    def warnings(self) -> List[Node]:
-        warnings.warn('Autosummary.warnings is deprecated',
-                      RemovedInSphinx40Warning, stacklevel=2)
-        return []
-
-    @property
-    def result(self) -> StringList:
-        warnings.warn('Autosummary.result is deprecated',
-                      RemovedInSphinx40Warning, stacklevel=2)
-        return self.bridge.result
-
 
 def strip_arg_typehint(s: str) -> str:
     """Strip a type hint from argument definition."""
@@ -487,8 +461,8 @@ def mangle_signature(sig: str, max_chars: int = 30) -> str:
         s = re.sub(r'{[^}]*}', '', s)
 
     # Parse the signature to arguments + options
-    args = []  # type: List[str]
-    opts = []  # type: List[str]
+    args: List[str] = []
+    opts: List[str] = []
 
     opt_re = re.compile(r"^(.*, |)([a-zA-Z0-9_*]+)\s*=\s*")
     while s:
@@ -605,7 +579,7 @@ def get_import_prefixes_from_env(env: BuildEnvironment) -> List[str]:
     Obtain current Python import prefixes (for `import_by_name`)
     from ``document.env``
     """
-    prefixes = [None]  # type: List[str]
+    prefixes: List[Optional[str]] = [None]
 
     currmodule = env.ref_context.get('py:module')
     if currmodule:
@@ -691,7 +665,7 @@ def import_ivar_by_name(name: str, prefixes: List[str] = [None]) -> Tuple[str, A
         analyzer = ModuleAnalyzer.for_module(modname)
         if (qualname, attr) in analyzer.find_attr_docs():
             return real_name + "." + attr, INSTANCEATTR, obj, modname
-    except (ImportError, ValueError):
+    except (ImportError, ValueError, PycodeError):
         pass
 
     raise ImportError
@@ -699,33 +673,6 @@ def import_ivar_by_name(name: str, prefixes: List[str] = [None]) -> Tuple[str, A
 
 # -- :autolink: (smart default role) -------------------------------------------
 
-def autolink_role(typ: str, rawtext: str, etext: str, lineno: int, inliner: Inliner,
-                  options: Dict = {}, content: List[str] = []
-                  ) -> Tuple[List[Node], List[system_message]]:
-    """Smart linking role.
-
-    Expands to ':obj:`text`' if `text` is an object that can be imported;
-    otherwise expands to '*text*'.
-    """
-    warnings.warn('autolink_role() is deprecated.', RemovedInSphinx40Warning, stacklevel=2)
-    env = inliner.document.settings.env
-    pyobj_role = env.get_domain('py').role('obj')
-    objects, msg = pyobj_role('obj', rawtext, etext, lineno, inliner, options, content)
-    if msg != []:
-        return objects, msg
-
-    assert len(objects) == 1
-    pending_xref = cast(addnodes.pending_xref, objects[0])
-    prefixes = get_import_prefixes_from_env(env)
-    try:
-        name, obj, parent, modname = import_by_name(pending_xref['reftarget'], prefixes)
-    except ImportError:
-        literal = cast(nodes.literal, pending_xref[0])
-        objects[0] = nodes.emphasis(rawtext, literal.astext(), classes=literal['classes'])
-
-    return objects, msg
-
-
 class AutoLink(SphinxRole):
     """Smart linking role.
 
@@ -760,7 +707,7 @@ def get_rst_suffix(app: Sphinx) -> str:
             return ('restructuredtext',)
         return parser_class.supported
 
-    suffix = None  # type: str
+    suffix: str = None
     for suffix in app.config.source_suffix:
         if 'restructuredtext' in get_supported_format(suffix):
             return suffix
@@ -826,7 +773,7 @@ def setup(app: Sphinx) -> Dict[str, Any]:
     app.connect('builder-inited', process_generate_options)
     app.add_config_value('autosummary_context', {}, True)
     app.add_config_value('autosummary_filename_map', {}, 'html')
-    app.add_config_value('autosummary_generate', [], True, [bool])
+    app.add_config_value('autosummary_generate', True, True, [bool])
     app.add_config_value('autosummary_generate_overwrite', True, False)
     app.add_config_value('autosummary_mock_imports',
                          lambda config: config.autodoc_mock_imports, 'env')
